﻿

namespace WindowsFormsGame15
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Text;
    using System.Windows.Forms;

    public class Game15
    {
        protected const int n = 4, m = n * n;
        Random rnd = new Random();
        /// <summary>
        /// values[i] - значение, хранящееся в ячейке i
        /// </summary>
        public int[] values = new int[m];
        /// <summary>
        /// indexes[k] - индекс ячейки в values, хранящей значение k
        /// indexes[values[i]] = i
        /// values[indexes[k]] = k
        /// </summary>
        public int[] indexes = new int[m];

        protected int G, S, E;    //Номера - цели, источника, пусто в диапазоне 0-15
        protected int r_g, c_g;   // индекс строки и столбца цели G
        protected int r_s, c_s;   // индекс строки и столбца источника S
        protected int r_e, c_e;   // индекс строки и столбца пустоты E

        protected string state;
        protected bool success;
        public string result;
        public int moves = 0;
        FormGame15 myForm_15;
        protected bool error = false;
        public bool exit = false;

        public Game15(FormGame15 form_game15)
        {
            myForm_15 = form_game15;
            Mix();
        }
        /// <summary>
        /// Создание новой случайной конфигурации
        /// </summary>
        public void Mix()
        {
            //чистка памяти
            for (int i = 0; i < m; i++)
                values[i] = indexes[i] = 0;
            int index;
            for (int i = 0; i < m - 1; i++)
            {
                index = rnd.Next(1, m);
                if (values[index] != 0)
                {//ячейка занята, ищем первую свободную
                    int k = (index + 1) % m;
                    while (values[k] != 0) k = (k + 1) % m; ;
                    index = k; //свободная ячейка обязательно найдется
                }
                values[index] = i + 1;
                indexes[i + 1] = index;
            }
            //Находим индекс свободной ячейки
            index = 0;
            while (values[index] != 0) index++;
            indexes[0] = index;
        }

        /// <summary>
        /// Построение конфигурации, заданной строкой текста
        /// Числа разделены пробелами. * - пустая ячейка.
        /// </summary>
        /// <param name="vals">входная строка</param>
        public bool GetFix(string vals)
        {
            string item = "";
            int kitem;
            int indexSpace_old = 0;
            int indexSpace = 0;
            try
            {
                for (int i = 0; i < n * n; i++)
                {
                    indexSpace = vals.IndexOf(' ', indexSpace_old);
                    item =
                        vals.Substring(indexSpace_old, indexSpace - indexSpace_old);
                    if (item == "*")
                        kitem = 0;
                    else
                        kitem = int.Parse(item);
                    values[i] = kitem;
                    indexes[kitem] = i;
                    indexSpace_old = indexSpace + 1;
                }
                for (int i = 0; i < n * n; i++)
                {
                    if (!values.Contains(i))
                        throw (new Exception());
                }
                return true;
            }
            catch (Exception)
            {
                result = "Загружаемая расстановка некорректна!";
                return false;
            }
        }

        /// <summary>
        /// Отладочная печать состояния конфигурации
        /// </summary>
        void Message()
        {
            string s1 = "values: ";
            string s2 = "indexes: ";
            for (int i = 0; i < m; i++)
            {
                s1 += values[i].ToString() + ", ";
                s2 += indexes[i].ToString() + ", ";
            }
            System.Windows.Forms.MessageBox.Show(s1 + "\r\n" + s2);
        }

        /// <summary>
        /// Инициализация переменных        
        /// </summary>        
        void Init()
        {
            E = indexes[0];
            S = indexes[G + 1];

            RC(E, ref r_e, ref c_e);
            RC(S, ref r_s, ref c_s);
            RC(G, ref r_g, ref c_g);
        }

        /// <summary>
        /// Пересчет координат
        /// </summary>
        public void RC(int p, ref int r, ref int c)
        {
            r = 1 + p / n; c = 1 + p % n;
        }

        /// <summary>
        /// Анализ состояния
        /// </summary>
        protected virtual void FindState()
        {
            if ((r_s - r_e) == 1 && c_s == c_e) state = "C1";  //Сосед сверху
            else if (r_e - r_s == 1 && c_s == c_e) state = "C2";  //Сосед снизу
            else
                //Сосед слева и S не самая правая точка
                if (c_s - c_e == 1 && r_s == r_e && c_s != n) state = "C4";
                else
                    if (c_e - c_s == 1 && r_s == r_e) state = "C3";  //Сосед справа
                    else
                        //Сосед слева и S самая правая точка
                        if (c_s - c_e == 1 && r_s == r_e) state = "C5";
                        else
                            //E не является соседом и E строго левее или выше S                    
                            if ((c_s - c_e > 1) && (r_s - r_e == 0) || (r_s - r_e >= 1)) 
                                state = "C6";
                            else
                                //E не является соседом и E правее или ниже S 
                                state = "C7"; //Возможно в начальном положении
        }

        /// <summary>
        /// Переход к новому состоянию 
        /// </summary>
        protected virtual void GoToState()
        {
            switch (state)
            {
                case "C1": C1(); break;
                case "C2": C2(); break;
                case "C3": C3(); break;
                case "C4": C4(); break;
                case "C5": C5(); break;
                case "C6": C6(); break;
                case "C7": C7(); break;
                default: throw (new Exception("Недопустимое состояние"));
            }
        }

        /// <summary>
        /// Запуск игры после инициализации
        /// </summary>
        public void StartGame()
        {
            moves = 0;
            exit = false;
            for (G = 0; G < (n - 1) * n; G++)
            {
                LocatePoint();
                //myForm_15.AddToProgressBar();
                myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
                if (exit)
                {
                    result = "Выполнение прервано";
                    return;
                }
            }
            LocateLastRow();
            // myForm_15.AddToResult();
            myForm_15.Invoke(new MethodInvoker(myForm_15.AddToResult));

        }

        /// <summary>
        /// Поставить на место точку G
        /// </summary>        
        void LocatePoint()
        {
            Init();
            FindState();
            //success = (G == S);
            success = EvalSuccess();
            while (!success)
                GoToState();

        }

        bool EvalSuccess()
        {
            bool success = true;
            for (int k = 0; k < G + 1; k++)
            {
                success = (values[k] == k + 1) && success;
            }
            return success;
        }


        /// <summary>
        /// упорядочить, если возможно последний ряд точек
        /// </summary>
        void LocateLastRow()
        {
            int i = E;
            while (i != 15) LeftMove(ref i);
            Finish(ref i);
            if (values[12] == 13 && values[13] == 15 && values[14] == 14
                ||
                values[12] == 14 && values[13] == 13 && values[14] == 15
                ||
                values[12] == 15 && values[13] == 14 && values[14] == 13)
                result = "Миссия невыполнима!";
            else
            {
                if (values[12] == 14 && values[13] == 15 && values[14] == 13)
                    FinalMoves_1();
                else
                    if (values[12] == 15 && values[13] == 13 && values[14] == 14)
                        FinalMoves_2();
                    else
                    {
                        //myForm_15.AddToProgressBar();
                        //myForm_15.AddToProgressBar();
                        // myForm_15.AddToProgressBar();
                        myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
                        myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
                        myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
                    }
                result = "Миссия выполнена!";
            }

        }
        void FinalMoves_1()
        {
            int i = 15;
            BottomMove(ref i); R3(ref i);
            TopMove(ref i); L3(ref i);
            BottomMove(ref i); RightMove(ref i); RightMove(ref i);
            TopMove(ref i); RightMove(ref i); BottomMove(ref i);
            L3(ref i); TopMove(ref i);
            R3(ref i); BottomMove(ref i);
            L3(ref i); TopMove(ref i);
            Finish(ref i);
            //myForm_15.AddToProgressBar(); 
            //myForm_15.AddToProgressBar();
            // myForm_15.AddToProgressBar();
            myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
            myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
            myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
        }
        void R3(ref int i)
        {
            RightMove(ref i); RightMove(ref i); RightMove(ref i);
        }
        void L3(ref int i)
        {
            LeftMove(ref i); LeftMove(ref i); LeftMove(ref i);
        }
        void FinalMoves_2()
        {
            int i = 15;
            BottomMove(ref i); R3(ref i);
            TopMove(ref i); L3(ref i);
            BottomMove(ref i); R3(ref i);
            TopMove(ref i); LeftMove(ref i); BottomMove(ref i);
            LeftMove(ref i); LeftMove(ref i); TopMove(ref i);
            R3(ref i); BottomMove(ref i);
            L3(ref i); TopMove(ref i);
            Finish(ref i);
            // myForm_15.AddToProgressBar();
            // myForm_15.AddToProgressBar();
            //myForm_15.AddToProgressBar();
            myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
            myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
            myForm_15.Invoke(new MethodInvoker(myForm_15.AddToProgressBar));
        }

        /// <summary>
        /// E является соседом сверху S. Цель G выше S. 
        /// </summary>
        protected void C1()
        {
            int i = E;
            //подъем вверх
            TopMove(ref i);

            Finish(ref i);
        }

        /// <summary>
        /// E является соседом снизу S. Цель G слева (и /или) выше S.        
        /// </summary>
        protected void C2()
        {
            if (G == S)
            {
                success = true; return;
            }
            int i = E;
            //цель слева  или S - самая правая-> сдвиг влево
            //переход в состояние "Пусто - левый сосед
            if (c_g < c_s || c_s == n)
            {
                RightMove(ref i);
                BottomMove(ref i);
                // LeftMove(ref i);
            }
            // Цель не слева и S -не самая правая-> сдвиг вправо
            // переход в состояние "Пусто - правый сосед
            else
            {
                LeftMove(ref i);
                BottomMove(ref i);
            }
            Finish(ref i);
        }

        /// <summary>
        /// E является соседом S справа. Цель G слева (и /или) выше S. 
        /// </summary>
        protected void C3()
        {

            if (G == S)
            {
                success = true; return;
            }
            int i = E;
            //Цель строго слева
            //переход в состояние "Пусто - левый сосед
            if (r_g == r_s && c_g < c_s)
            {//обход S (подъем в Е, вправо, вниз               
                TopMove(ref i);
                RightMove(ref i);
                RightMove(ref i);
                BottomMove(ref i);
                LeftMove(ref i);

            }
            //цель справа -> сдвиг вправо
            else if (c_g > c_s)
            {
                RightMove(ref i);

            }

            //цель строго выше или выше и левее -> обход и подъем 
            else
            {
                BottomMove(ref i);
                RightMove(ref i);
                TopMove(ref i);

            }
            Finish(ref i);
        }

        /// <summary>
        /// E является соседом S слева. Цель G слева (и /или) выше S.
        ///  S не самая правая
        /// </summary>
        protected void C4()
        {
            int i = E;
            //цель строго слева
            if (r_g == r_s && c_g < c_s)
            {
                LeftMove(ref i);
            }
            //иначе цель выше
            //цель левее или не в следующем ряду -> возможен левый обход
            else if (c_g < c_s || r_s - r_g > 1) //левый обход
            {
                BottomMove(ref i);
                LeftMove(ref i);
                TopMove(ref i);
            }
            // цель справа и S не в последнем ряду ->обход и сдвиг вправо
            else if (c_g > c_s && r_s != n)
            {
                TopMove(ref i);
                LeftMove(ref i);
                LeftMove(ref i);
                BottomMove(ref i);
                RightMove(ref i);

            }
            //цель строго выше в следующем ряду и S не в последнем ряду-> правый обход
            else if (c_g == c_s && r_s - r_g == 1 && r_s != n)
            {
                TopMove(ref i);
                LeftMove(ref i);
                LeftMove(ref i);
                BottomMove(ref i);
                BottomMove(ref i);
                RightMove(ref i);
                TopMove(ref i);
            }
            else
                //цель строго выше и источник в последнем ряду - > сдвиги
                if (c_g == c_s && r_s - r_g == 1 && r_s == n)
                {

                    BottomMove(ref i);
                    for (int k = c_s; k < n; k++)
                        LeftMove(ref i);
                    TopMove(ref i);
                    RightMove(ref i);
                    BottomMove(ref i);
                    RightMove(ref i);
                    TopMove(ref i);
                }
                //цель справа в следующем ряду и S в последнем ряду  -> сдвиг вправо
                else
                    if (c_g > c_s && r_s == n)
                    {
                        BottomMove(ref i);
                        for (int k = 0; k < c_g - c_s; k++)
                            LeftMove(ref i);
                        TopMove(ref i);
                        if (S == i)
                        {
                            LeftMove(ref i);
                            BottomMove(ref i);
                            for (int k = 0; k <= c_g - c_s; k++)
                                RightMove(ref i);
                            TopMove(ref i);
                        }
                        else // сдвиг вправо
                        {
                            for (int k = 0; k < c_g - c_s; k++)
                                RightMove(ref i);
                            BottomMove(ref i);
                            for (int k = 0; k < c_g - c_s; k++)
                                LeftMove(ref i);
                            TopMove(ref i);
                            LeftMove(ref i);
                            BottomMove(ref i);
                            RightMove(ref i);
                            for (int k = 0; k < c_g - c_s; k++)
                                RightMove(ref i);
                            TopMove(ref i);
                            for (int k = 0; k < c_g - c_s; k++)
                                LeftMove(ref i);
                            BottomMove(ref i);
                            for (int k = 0; k < c_g - c_s; k++)
                                RightMove(ref i);
                            TopMove(ref i);


                        }

                    }
                    else throw (new Exception());
            Finish(ref i);
        }

        /// <summary>
        /// E является соседом слева S. S -самая правая.
        /// </summary>
        protected void C5()
        {
            int i = E;
            //цель строго  слева -> сдвиг влево
            if (r_g == r_s && c_g < c_s)
            {
                LeftMove(ref i);
                state = "C3";
            }
            //Цель строго выше в следующем ряду -> сдвиг ряда 
            else if (r_s - r_g == 1 && c_s == c_g)
            {
                RightMove(ref i);
                RightMove(ref i);
                BottomMove(ref i);
                LeftMove(ref i);
                LeftMove(ref i);
                LeftMove(ref i);
                TopMove(ref i);
                RightMove(ref i);
                BottomMove(ref i);
                RightMove(ref i);
                RightMove(ref i);
                TopMove(ref i);
            }
            else
            {
                BottomMove(ref i);
                LeftMove(ref i);
                TopMove(ref i);
                state = "C2";
            }
            Finish(ref i);
        }

        /// <summary>
        /// E не является соседом S.
        /// E строго слева (и /или) выше S
        /// </summary>
        protected void C6()
        {
            if (G == S)
            {
                success = true; return;
            }
            //Проходим путь от S к E
            int i = E;
            //Опускаем вниз пусто. Это всегда возможно
            for (int j = r_e; j < r_s; j++) //перемещение вверх
            {
                TopMove(ref i);
            }
            //движемся влево или вправо
            if (c_s > c_e)
                for (int j = c_e; j < c_s; j++) //перемещение влево
                {
                    LeftMove(ref i);
                }
            else
                for (int j = c_s; j < c_e; j++) //перемещение вправо
                {
                    RightMove(ref i);
                }
            Finish(ref i);
        }
        /// <summary>
        /// Путь влево от источника к цели
        /// </summary>
        void LeftSG()
        {
            for (int i = G; i < S; i++)
            {
                values[i] = values[i + 1];
                indexes[values[i + 1]] -= 1;
            }
            indexes[0] = S;
            values[S] = 0;
            Init();
            if (S == G) success = true;
            state = "C3";
        }
        /// <summary>
        /// E не является соседом S. Цель G слева (и /или) выше S.
        /// E не совпадает с целью. E слева (и /или) выше S
        /// </summary>
        void C7()
        {
            if (G == S)
            {
                success = true; return;
            }
            int i = E;
            if(c_s > c_e)
            for (int k = 0; k < c_s - c_e -1; k++)
                 LeftMove(ref i);            
            else
                for (int k = 0; k < c_e - c_s -1; k++)
                    RightMove(ref i);
            if (r_s > r_e)
                for (int k = 0; k < r_s - r_e; k++)
                    TopMove(ref i);
            else
                for (int k = 0; k < r_e - r_s; k++)
                    BottomMove(ref i);
                

            Finish(ref i);
        }


        /// <summary>
        /// Ход слева в ячейку i
        /// </summary>
        /// <param name="i"> цель</param>
        protected void LeftMove(ref int i)
        {
            values[i] = values[i + 1];
            indexes[values[i + 1]] -= 1;
            i = i + 1;
            moves++;
            // myForm_15.Moving(values[i]);
            myForm_15.Invoke(myForm_15.delint, new object[] { values[i] });
            ToList(i, i - 1);
        }
        /// <summary>
        /// Ход справа в ячейку i
        /// </summary>
        /// <param name="i">цель</param>
        protected void RightMove(ref int i)
        {
            values[i] = values[i - 1];
            indexes[values[i - 1]] += 1;
            i = i - 1;
            moves++;
            //myForm_15.Moving(values[i]);
            myForm_15.Invoke(myForm_15.delint, new object[] { values[i] });
            ToList(i, i + 1);
        }

        /// <summary>
        /// Ход вверх в ячейку i
        /// </summary>
        /// <param name="i">цель</param>
        protected void TopMove(ref int i)
        {
            values[i] = values[i + n];
            indexes[values[i + n]] -= n;
            i = i + n;
            moves++;
            //myForm_15.Moving(values[i]);
            myForm_15.Invoke(myForm_15.delint, new object[] { values[i] });
            ToList(i, i - n);
        }
        /// <summary>
        /// Ход вниз в ячейку i
        /// </summary>
        /// <param name="i">цель</param>
        protected void BottomMove(ref int i)
        {
            values[i] = values[i - n];
            indexes[values[i - n]] += n;
            i = i - n;
            moves++;
            //myForm_15.Moving(values[i]);
            myForm_15.Invoke(myForm_15.delint, new object[] { values[i] });
            ToList(i, i + n);
        }

        /// <summary>
        /// Формирование строки, задающей ход
        /// </summary>
        /// <param name="i"></param>
        /// <param name="j"></param>
        void ToList(int i, int j)
        {
            string move = moves.ToString() + ". ";

            string name = "";
            name += GetName(i);
            move += name + " -> ";
            move += GetName(j);
            //myForm_15.ToList(move);
            myForm_15.Invoke(myForm_15.delstring, new object[] { move });
        }

        string GetName(int i)
        {
            int r = 0, c = 0;
            string move = "";
            RC(i, ref r, ref c);
            switch (c)
            {
                case 1: move += "A"; break;
                case 2: move += "B"; break;
                case 3: move += "C"; break;
                case 4: move += "D"; break;
                default: System.Windows.Forms.MessageBox.Show("Error - move!"); break;
            }
            switch (r)
            {
                case 1: move += "1"; break;
                case 2: move += "2"; break;
                case 3: move += "3"; break;
                case 4: move += "4"; break;
                default: System.Windows.Forms.MessageBox.Show("Error - move!"); break;
            }
            return (move);

        }
        /// <summary>
        /// Завершение кофигурации
        /// </summary>
        protected void Finish(ref int i)
        {
            indexes[0] = i;
            values[i] = 0;
            Init();
            if (S == G) success = true;
            FindState();
        }
    }
}
