﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ClassLibraryIntegral
{
    
    public delegate double Integral_function(double x);
    public class Integral
    {
        double a, b; //пределы интегрирования
        Integral_function f; //подынтегральная функция
        int p; //число сегментов разбиения
        double eps; //точность вычисления
        double result; //результат вычисления
        object locker = new object();
        double[] results;

        public double Result
        {
        get { return result; }
        }

        /// <summary>
        /// конструктор
        /// </summary>
        /// <param name="a">нижний предел интегрирования</param>
        /// <param name="b">верхний предел интегрирования</param>
        /// <param name="p">число сегментов разбиения интервала
        /// интегрирования</param>
        /// <param name="f">подынтегральная функция</param>
        /// <param name="eps">точность вычисления
        /// интеграла</param>
        public Integral(double a, double b, int p,
                            Integral_function f, double eps)
        {
            this.a = a;
            this.b = b;
            this.p = p;
            this.f = f;
            this.eps = eps;
            results = new double[p];
        }

        /// <summary>
        /// Последовательный алгоритм
        /// </summary>
        /// <param name="a">начало отрезка интегрирования</param>
        /// <param name="b">конец отрезка интегрирования</param>
        /// <param name="result" значение интеграла </param>        /// 
        void DefiniteIntegral(double a, double b, out double result)
        {
            int n = 2;
            double dx = (b - a) / 2;
            double S0 = 0, S = 0;
            double x = 0;
            bool success = false;
            for (int i = 0; i < n; i++)
            {
                x = a + i * dx;
                S0 += f(x);
            }
            S0 *= dx;
            while (!success)
            {
                n = 2 * n;
                dx = dx / 2;
                S = 0;
                for (int i = 1; i < n; i += 2)
                {
                    x = a + i * dx;
                    S += f(x);
                }
                S = S * dx + S0 / 2;
                if (Math.Abs(S - S0) > eps)
                    S0 = S;
                else
                    success = true;
            }
            result = S;
        }

        /// <summary>
        /// Последовательный алгоритм
        /// Метод трапеций
        /// </summary>
        /// <param name="a">начало отрезка интегрирования</param>
        /// <param name="b">конец отрезка интегрирования</param>
        /// <param name="result" значение интеграла </param>        /// 
        void DefiniteIntegralT(double a, double b, out double result)
        {
            int n = 1;
            double dx0 = (b - a);
            double dx = dx0;
            double Sum0 = (f(a) + f(b)) / 2;
            double S0 = Sum0 * dx0;
            double Sum1 = Sum0;
            double S1 = S0;
            double x0 = a;
            double x = x0;
           
            bool success = false;
            int i = 1;
            while (!success)
            {                
                dx = dx / 2;
                Sum1 = 0;
                x0 = a + dx;
                for (int k = 0; k < n; k++)
                {
                    x = x0 + k * dx0;
                    Sum1 += f(x);
                }
                Sum1 += Sum0;
                S1 = Sum1 * dx;
                dx0 = dx;
                n *= 2;
                i++;
                if (Math.Abs(S1 - S0) > eps)
                {

                    Sum0 = Sum1;
                    S0 = S1;
                }
                else
                    success = true;
            }
            result = S1;
        }
        public double SequenceIntegral()
        {
            double result;
            //DefiniteIntegral(a, b, out result);
            DefiniteIntegralT(a, b, out result);
            return result;
        }

        /// <summary>
        /// Последовательный алгоритм
        /// Вычисление с разбиением интервала интегрирования
        /// на p сегментов
        /// </summary>
        public double SequenceIntegralWithSegments(int p)
        {
            results = new double[p];
            
            double dx = (b - a) / p;
            double start = 0, finish = 0;
            for (int i = 0; i < p; i++)
            {
                start = a + i * dx;
                finish = start + dx;
               // DefiniteIntegral(start,
               // finish, out results[i]);
                DefiniteIntegralT(start,
               finish, out results[i]);
            }
            result = 0;
            for (int i = 0; i < p; i++)
            {
                result += results[i];
            }
            return result;
        }

        /// <summary>
        /// Параллельный алгоритм
        /// Вычисление с использованием Parallel.For
        /// </summary>
        public double ParallelIntegralWithSegments(int p)
        {
            results = new double[p];
            
            Parallel.For(0, p,  (i) =>IntegralP(i, p));
            {
               
            }
            result = 0;
            for (int j = 0; j < p; j++)
            {
                result += results[j];
            }
            return result;
        }
        void IntegralP(int i, int p)
        {
            double dx = (b - a) / p;
            double start = 0, finish = 0;
            start = a + i * dx;
            finish = start + dx;
           // DefiniteIntegral(start,
           //     finish, out results[i]);
            DefiniteIntegralT(start,
                finish, out results[i]);
        }
    }
}
