﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;


namespace ClassLibrarySorting
{
    public static class Sorting
    {
       static Random rnd = new Random();
       /// <summary>
       /// Создание массива вещественных чисел из диапазона -
       /// [-max ... max]
       /// со случайными значениями
       /// </summary>
       /// <param name="n">размерность массива</param>
       /// <param name="max">диапазон значений</param>
       /// <returns>массив случайных чисел</returns>
       public static double[] Create_Random_Massiv(int n, int max)
       {
           double[] mas = new double[n];
           for (int i = 0; i < n; i++)
               mas[i] = (rnd.NextDouble() - 0.5) * 2 * max;

           return mas;
       }

       public static bool Is_Sorted(double[] mas)
       {
           int n = mas.Length;
           bool result = true;
           for (int i = 0; i < n -1; i++)
           {
                if(mas[i] > mas[i+1])
                { result = false; break; }
           }
           return result;

       }

       /// <summary>
       /// Клаcсический вариант пузырьковой сортировки
       /// Со сложностью O(n * n)
       /// </summary>
       /// <param name="mas">сортируемый массив</param>
       public static void BubbleSortClassic(double[] mas)
       {
           int n = mas.Length;
           double temp = 0;
           for (int k = 0; k < n - 1; k++)
           { //цикл по числу проходов               
               for (int i = n - 1; i > k; i--)
               {  // цикл всплытия
                   if (mas[i] < mas[i - 1])
                   {//swap
                       temp = mas[i];
                       mas[i] = mas[i - 1];
                       mas[i - 1] = temp;
                   }
               }
           }
       }


       /// <summary>
       /// Параллельный вариант пузырьковой сортировки         
       /// </summary>
       /// <param name="mas">сортируемый массив</param>
       /// <param name="p">число групп</param>
       public static void BubbleParallel(double[] mas, int p)
       {
           
           //Цикл по числу групп
           Parallel.For(0, p, (j) =>
               { Sort_Group(mas, p, j); });
           //Слияние отсортированных последовательностей
           Merge(mas, p);
       }
           
    static void Sort_Group( double[] mas, int p, int j)
    {
        // процессор j сортирует подпоследовательность элементов, 
        //начинающихся индексом n -j -1 и отстоящих на расстоянии р
        //цикл по числу проходов m
        int n = mas.Length;
        int i0 = n - j - 1;
        int m = i0 / p;
        double temp = 0;
        for (int k = 0; k < m; k++)
        {
            //цикл всплытия легкого элемента на k-м проходе            
            for (int i = i0; i - p >= k * p; i = i - p)
            {
                if (mas[i] < mas[i - p])
                {//swap
                    temp = mas[i];
                    mas[i] = mas[i - p];
                    mas[i - p] = temp;
                }
            }
        }
    }
           

       /// <summary>
       /// Слияние упорядоченных последовательностей
       /// Последовательности представляют отрезки с шагом p
       /// Используется дополнительная память
       /// </summary>
       /// <param name="mas">сортируемый массив</param>
       /// <param name="p">число процессоров</param>
       static void Merge(double[] mas, int p)
       {
           int n = mas.Length;
           int m = n / p;
           int index_min = 0;
           double min = 0;
           int i = 0;
           double[] tmas = new double[n];
           int[] start = new int[p], finish = new int[p];
           for (i = 1; i <= p; i++)
           {
               finish[p - i] = n - i;
               start[p - i] = finish[p - i] % p;
           }
           for (int k = 0; k < n; k++)
           {//пересылка k-ого элемента
               //поиск кандидата
               i = 0;
               while (start[i] > finish[i]) i++;
               index_min = i; min = mas[start[i]];

               for (int j = i + 1; j < p; j++)
               {
                   //цикл по кандидатам
                   if (start[j] <= finish[j])
                   {
                       if (mas[start[j]] < min)
                       {
                           min = mas[start[j]];
                           index_min = j;
                       }
                   }
               }
               //pass                        
               tmas[k] = mas[start[index_min]];
               start[index_min] += p;
           }
           for (i = 0; i < n; i++)
               mas[i] = tmas[i];
       }

       /// <summary>
       /// Быстрая сортировка Хоара
       /// Вызов рекурсивной версии
       /// </summary>
       /// <param name="mas"></param>
       public static void QuickSort(double[] mas)
       {
           QSort(mas, 0, mas.Length - 1);
       }
       /// <summary>
       /// Быстрая сортировка Хоара
       /// Вызов рекурсивной параллельной версии
       /// </summary>
       /// <param name="mas"></param>
       public static void PQuickSort(double[] mas)
       {
           PQSort(mas, 0, mas.Length - 1);
       }

       /// <summary>
       /// Рекурсивная версия быстрой сортировки
       /// </summary>
       /// <param name="mas">сортируемый массив</param>
       /// <param name="start">индекс первого элемента сортируемой части</param>
       /// <param name="finish">индекс последнего элемента сортируемой части массива</param>
       static void QSort(double[] mas, int start, int finish)
       {
           
           if (finish - start > 0)
           {
               double cand = 0, temp = 0;
               int l = start, r = finish;
               cand = mas[(r + l) / 2];
               while (l <= r)
               {
                   while (mas[l] < cand) l++;
                   while (mas[r] > cand) r--;
                   if (l <= r)
                   {
                       temp = mas[l];
                       mas[l] = mas[r];
                       mas[r] = temp;
                       l++; r--;
                   }
               }
               QSort(mas, start, r);
               QSort(mas, l, finish);
           }
       }

       /// <summary>
       /// Быстрая сортировка Хоара
       /// Параллельная версия
       /// </summary>
       /// <param name="mas"></param>
       /// <param name="p">число процессоров</param>
       public static void QuickSortParallel(double[] mas, int p)
       {
           int n = mas.Length;
           int m = n / p;

           Parallel.For(0, p, (i) =>
           {
               int start = i * m;
               int finish = i != p - 1 ? start + m - 1 : n - 1;
               QSort(mas, start, finish);
           });

           MergeQ(mas, p);
       }

       /// <summary>
       /// Слияние упорядоченных последовательностей
       /// Последовательности представляют подряд идущие отрезки
       /// Используется дополнительная память
       /// </summary>
       /// <param name="mas">сортируемый массив</param>
       /// <param name="p">число групп</param>
       public static void MergeQ(double[] mas, int p)
       {
           int n = mas.Length;
           int m = n / p;
           int index_min = 0;
           double min = 0;
           int i = 0;
           double[] tmas = new double[n];
           int[] start = new int[p], finish = new int[p];
           for (i = 0; i < p; i++)
           {
               start[i] = i * m;
               finish[i] = i != p - 1 ? start[i] + m - 1 : n - 1;
           }
           for (int k = 0; k < n; k++)
           {//пересылка k-ого элемента
               //поиск кандидата
               i = 0;
               while (start[i] > finish[i]) i++;
               index_min = i; min = mas[start[i]];

               for (int j = i + 1; j < p; j++)
               {
                   //цикл по кандидатам
                   if (start[j] <= finish[j])
                   {
                       if (mas[start[j]] < min)
                       {
                           min = mas[start[j]];
                           index_min = j;
                       }
                   }
               }
               //pass                        
               tmas[k] = mas[start[index_min]];
               start[index_min]++;
           }
           Array.Copy(tmas, mas, n);
       }
        
       /// <summary>
       /// Рекурсивная параллельная версия быстрой сортировки
       /// </summary>
       /// <param name="mas">сортируемый массив</param>
       /// <param name="start">индекс первого элемента сортируемой части</param>
       /// <param name="finish">индекс последнего элемента сортируемой части массива</param>
       static void PQSort(double[] mas, int start, int finish)
       {
           const int limit = 100000;
           Task task1, task2;
           if (finish - start > 0)
           {
               double cand = 0, temp = 0;
               int l = start, r = finish;
               cand = mas[(r + l) / 2];
               while (l <= r)
               {
                   while (mas[l] < cand) l++;
                   while (mas[r] > cand) r--;
                   if (l <= r)
                   {
                       temp = mas[l];
                       mas[l] = mas[r];
                       mas[r] = temp;
                       l++; r--;
                   }
               }
               bool p1, p2;
               p1 = (r - start) > limit;
               p2 = (finish - l) > limit;

               if (p1)
               {
                   task1 = new Task(() =>
                       { PQSort(mas, start, r); });
                   task1.Start();
               }
               else
               {
                   task1 = new Task(() => { });
                   QSort(mas, start, r);
               }
               if (p2)
               {
                   task2 = new Task(() =>
                        { PQSort(mas, l, finish); });
                   task2.Start();
               }
               else
               {
                   task2 = new Task(() => { });
                   QSort(mas, l, finish);
               }
               if (p1) task1.Wait();
               if (p2) task2.Wait();
           }
       }
       
      /// <summary>
      /// Сортировка Бэтчера
      /// Последовательная версия
      /// Алгоритм, описанный Кнутом
      /// Внимание !!!
      /// Пригоден только для N = 2 ^ K
      /// </summary>
      /// <param name="mas"></param>  
      public static void Betcher(double[] mas)
       {
           int n = mas.Length;
           int t = (int)Math.Log(n + 1, 2);
           int p = (int)Math.Pow(2, t - 1);
           int q = 0;
           int r = 0, d = 0;
           double temp = 0;
           int t1 = 0, t2 = 0;

           while (p != 0)
           {
               r = 0;
               d = p;
               q = (int)Math.Pow(2, t-1);
               while (q >= p)
               {
                   for (int i = 0; i < n - d; i++)
                   {
                       t1 = i; t2 = t1 +d;
                       if (((i & p) == r) && (mas[t1] > mas[t2]))
                       {
                           //swap                           
                           temp = mas[t1];
                           mas[t1] = mas[t2];
                           mas[t2] = temp;
                       }
                   }
                   d = q - p;
                   q = q / 2;
                   r = p;
               }
               p = p / 2;
           }
       }

      /// <summary>
      /// Сортировка Бэтчера
      /// Параллельная версия
      ///  Алгоритм, описанный Кнутом
      /// Внимание !!!
      /// Пригоден только для N = 2 ^ K
      /// </summary>
      /// <param name="mas"></param>  
      public static void PBetcher(double[] mas)
      {
          int n = mas.Length;
          int t = (int)Math.Log(n + 1, 2);
          int p = (int)Math.Pow(2, t - 1);
          int q = 0;
          int r = 0, d = 0;
          
          while (p != 0)
          {
              r = 0;
              d = p;
              q = (int)Math.Pow(2, t - 1);
              while (q >= p)
              {
                  Parallel.For(0, n - d, (i) => Sorter(mas, i, p, d, r));                  
                  d = q - p;
                  q = q / 2;
                  r = p;
              }
              p = p / 2;
          }
      }
      static void Sorter(double[] mas, int i, int p, int d, int r)
      {
          double temp = 0;
          int t1 = 0, t2 = 0; 
          t1 = i; t2 = t1 + d;
          if (((i & p) == r) && (mas[t1] > mas[t2]))
          {
              //swap                           
              temp = mas[t1];
              mas[t1] = mas[t2];
              mas[t2] = temp;
          }
      }
      static void compexch(ref double a, ref double b)
      {
          double temp = 0;
          if (a > b)
          {
              temp = a;
              a = b;
              b = temp;
          }
      }

      /// <summary>
      /// Сортировка Бэтчера
      /// Последовательный вариант по книге Седжвика
      /// Пригоден для любых N
      /// </summary>
      /// <param name="a"></param>
      public static void Batchersort(double[] mas)
      {
          int i, j, k, p, N = mas.Length;
          for (p = 1; p < N; p += p)
              for (k = p; k > 0; k /= 2)
                  for (j = k % p; j + k < N; j += (k + k))
                      for (i = 0; i < k; i++)
                          if (j + i + k < N)
                              if ((j + i) / (p + p) == (j + i + k) / (p + p))
                                  compexch(ref mas[j + i], ref mas[j + i + k]);
      }

      /// <summary>
      /// Параллельная Сортировка Бэтчера
      /// </summary>
      /// <param name="a"></param>
      public static void PBatchersort(double[] a)
      {
          int j, k, p, N = a.Length;
          
          for (p = 1; p < N; p += p)
              for (k = p; k > 0; k /= 2)
                  for (j = k % p; j + k < N; j += (k + k))
                      Parallel.For(0, k, (i) => SortGroup(a, i, j, k, p));
                          
      }
      static void SortGroup(double[] a, int i, int j, int k, int p)
      {
          int N = a.Length;
          int t1 = 0, t2 = 0, t3 = 0;
          t1 = i + j; t2 = t1 + k; t3 = p + p;
          if (t2 < N)
              if (t1 / t3 == t2 / t3)
                  compexch(ref a[t1], ref a[t2]);
      }

    }
}
