﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;

namespace ClassLibrarySum
{
    /// <summary>
    /// Суммирование бесконечного сходящегося ряда   
    /// </summary>
    public class Infinite_series
    {
        double[] ar;    //массив
        int n; //Размер массива
        int p;  //Число групп при делении массива
        int m; //Число элементов в группе
        double eps = 1e-13;
        double[] y;     // рабочий массив
        double x2p = 0; //рабочая переменная
        //Вычисление суммы сходящегося ряда на примере 
        //функции Arcsin(x)

        /// <summary>
        /// Конструктор класса, 
        /// которому передается точность вычисления функции
        /// </summary>
        /// <param name="eps"></param>
        public Infinite_series(double eps)
        {
            this.eps = eps;
        }

        /// <summary>
        /// Последовательный алгоритм
        /// </summary>
        /// <param name="x">аргумент</param>
        /// <returns>Значение функции Arcsin</returns>
        public double Seq_ArcSin(double x)
        {
            double x2 = x * x;
            int i = 0;
            double a = x;   //начальное значение 
            int k = 0;
            double S = 0;
            a = x; k = 0; i = 0;
            S = a;
            while (Math.Abs(a) > eps)
            {
                k = 2 * i + 1;
                a *= k * k * x2 / ((k + 1) * (k + 2));
                S += a;
                i++;
            }
            return S;
        }

        /// <summary>
        /// Параллельный шаговый алгоритм
        /// </summary>
        /// <param name="x">аргумент функции</param>
        /// <param name="p">число групп суммирования</param>
        /// <returns>Значение функции Arcsin</returns>
        public double Parallel_ArcSin(double x, int p)
        {
            double x2 = x * x;
            double i = 0;
            double k = 0;
            y = new double[p];
            x2p = x2;
            double S = 0;
            //Вычисление начальных значений 
            y[0] = x;
            for (int j = 1; j < p; j++)
            {
                k = 2 * i + 1;
                y[j] = y[j - 1] * k * k * x2 / ((k + 1) * (k + 2));
                i++;
                x2p *= x2;
            }
            //Параллельное вычисление сумм для каждой группы
            Parallel.For(0, p, (j) => { Summa_Asin(j, x2p, p); });

            //Суммирование результатов групп
            S = y[0];
            for (int j = 1; j < p; j++)
                S += y[j];
            return S;
        }

        /// <summary>
        /// Метод, выполняемый на итерациях Parallel.For
        /// Результаты запоминаются в глобальном массиве y
        /// </summary>
        /// <param name="j">Индекс итерации</param>
        /// <param name="x2p">значение степени, 
        /// вычисленное до начала цикла</param>
        /// <param name="p">номер группы</param>
        void Summa_Asin(int j, double x2p, int p)
        {
            double a = 0;
            a = y[j];
            int i = j;
            double pr = 1;

            while (Math.Abs(a) > eps)
            {
                pr = 1;
                for (int r = 1; r <= p; r++)
                    pr *= (1 - 1.0 / (2 * (i + r)));
                a *= pr * (2 * i + 1) / ((2 * (i + p) + 1)) * x2p;
                y[j] += a;
                i += p;
            }
        }

    }
}

