﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ClassLibrarySum
{
    /// <summary>
    /// последовательный и параллельные методы
    /// вычисления суммы элементов массива
    /// </summary>
    public class Massiv
    {
        double[] ar;    //массив
        int n; //Размер массива
        int p;  //Число групп при делении массива
        int m; //Число элементов в группе        
        double[] y;     // рабочий массив        
        int Range = 1000;   //диапазон элементов
        Random rnd = new Random();

       
        /// <summary>
        /// конструктор
        /// </summary>
        /// <param name="size"></param>
        /// <param name="p"></param>
        /// <param name="Range"></param>
        public Massiv(int size, int p, int Range)
        {
            n = size;
            this.p = p;
            this.Range = Range;
            ar = new double[n];
            y = new double[n];

        }
        /// <summary>
        /// Заполнение массива
        /// случайными числами из заданного диапазона
        /// </summary>
        public void CreateRandomMassiv()
        {
            for (int i = 0; i < n; i++)
                ar[i] = (rnd.NextDouble() - 0.5) * Range;
        }

        /// <summary>
        /// Классический последовательный алгоритм,
        /// использующий оператор цикла for
        /// </summary>
        /// <param name="time">время работы</param>
        /// <returns>сумма элементов массива</returns>
        public double Sum_Seq(out double time)
        {
            DateTime start, finish;
            start = DateTime.Now;
            double S = 0;
            for (int i = 0; i < n; i++)
                S += ar[i];
            finish = DateTime.Now;
            time = (finish - start).Ticks * 1E-7;
            return S;
        }

        /// <summary>
        /// Параллельный пирамидальный алгоритм
        /// вычисления суммы элементов массива
        /// Реализация использует метод Parallel.For
        /// и анонимный метод для вычисления пирамиды
        /// </summary>
        /// <param name="time">время работы</param>
        /// <returns>сумма элементов массива</returns>
        public double Sum_Piramid(out double time)
        {
            DateTime start, finish;
            Array.Copy(ar, y, n);
            start = DateTime.Now;
            double S = 0;
            m = n;
            while (m != 1)
            {
                Parallel.For(0, m / 2, (i) =>
                {
                    y[i] = y[i] + y[m - i - 1];
                });
                m = (m + 1) / 2;
            }
            S = y[0];
            finish = DateTime.Now;
            time = (finish - start).Ticks * 1E-7;
            return S;
        }
        /// <summary>
        /// Другая реализация пирамидального метода
        /// Классическая схема метода Parallel.For,
        /// не использующая анонимный метод
        /// </summary>
        /// <param name="time">время работы</param>
        /// <returns>сумма элементов массива</returns>
        public double Summa_Piramid(out double time)
        {
            DateTime start, finish;
            Array.Copy(ar, y, n);
            start = DateTime.Now;
            double S = 0;
            m = n;
            while (m != 1)
            {
                Parallel.For(0, m / 2, Summa);
                m = (m + 1) / 2;
            }
            S = y[0];
            finish = DateTime.Now;
            time = (finish - start).Ticks * 1E-7;
            return S;
        }
        void Summa(int i)
        {
            y[i] = y[i] + y[m - i - 1];
        }

        /// <summary>
        /// Параллельный сегментный алгоритм
        /// вычисления суммы элементов массива
        /// Неиспользуемая реализация!!!
        /// Реализация использует метод Parallel.For
        /// и анонимный метод
        /// Пример некорректной работы анонимного метода!!!
        /// </summary>
        /// <param name="time">время работы</param>
        /// <returns>сумма элементов массива</returns>
        public double Sum_Segment(out double time)
        {
            DateTime start, finish;
            y = new double[p];
            m = n / p + 1;
            int st = 0, fin = 0;
            start = DateTime.Now;
            double S = 0;
            Parallel.For(0, p, (k) =>
            {
                st = k * m;
                fin = (k + 1) * m < n ? (k + 1) * m : n;
                for (int i = st; i < fin; i++)
                    y[k] += ar[i];
            });
            S = y[0];
            for (int i = 1; i < p; i++)
                S += y[i];
            finish = DateTime.Now;
            time = (finish - start).Ticks * 1E-7;
            return S;
        }

        /// <summary>
        /// Параллельный сегментный алгоритм
        /// вычисления суммы элементов массива
        /// Реализация использует метод Parallel.For
        /// в классическом варианте без использования 
        /// анонимного метода
        /// </summary>
        /// <param name="time">время работы</param>
        /// <returns>сумма элементов массива</returns>
        public double Sum_Seg(out double time)
        {
            DateTime start, finish;
            y = new double[p];
            m = n / p + 1;

            start = DateTime.Now;
            double S = 0;
            Parallel.For(0, p, Summa_Seg);

            S = y[0];
            for (int i = 1; i < p; i++)
                S += y[i];
            finish = DateTime.Now;
            time = (finish - start).Ticks * 1E-7;
            return S;
        }
        void Summa_Seg(int k)
        {
            int start = 0, finish = 0;
            start = k * m;
            finish = (k + 1) * m < n ? (k + 1) * m : n;
            for (int i = start; i < finish; i++)
                y[k] += ar[i];
        }

        /// <summary>
        /// Параллельный шаговый алгоритм
        /// вычисления суммы элементов массива
        /// Реализация использует метод Parallel.For
        /// в классическом варианте 
        /// без использования анонимного метода
        /// </summary>
        /// <param name="time">время работы</param>
        /// <returns>сумма элементов массива</returns>
        public double Sum_Step(out double time)
        {
            DateTime start, finish;
            y = new double[p];

            start = DateTime.Now;
            double S = 0;
            Parallel.For(0, p, Summa_Step);

            S = y[0];
            for (int i = 1; i < p; i++)
                S += y[i];
            finish = DateTime.Now;
            time = (finish - start).Ticks * 1E-7;
            return S;
        }
        void Summa_Step(int k)
        {

            for (int i = k; i < n; i += p)
                y[k] += ar[i];
        }

        /// <summary>
        /// Параллельный шаговый алгоритм
        /// вычисления суммы элементов массива
        /// Реализация использует явную работу
        /// с потоками
        /// </summary>
        /// <param name="time">время работы</param>
        /// <returns>сумма элементов массива</returns>
        public double Sum_Thread(out double time)
        {
            DateTime start, finish;
            y = new double[p];
            Thread[] threads = new Thread[p];

            for (int i = 0; i < p; i++)
                threads[i] = new Thread(Summa_Thread);
            start = DateTime.Now;
            double S = 0;
            for (int i = 0; i < p; i++)
                threads[i].Start(i);
            for (int i = 0; i < p; i++)
                threads[i].Join();
            S = y[0];
            for (int i = 1; i < p; i++)
                S += y[i];
            finish = DateTime.Now;
            time = (finish - start).Ticks * 1E-7;
            return S;
        }
        void Summa_Thread(object ind)
        {
            int k = (int)ind;
            for (int i = k; i < n; i += p)
                y[k] += ar[i];
        }
    }
}
