﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ClassLibrarySum
{
    /// <summary>
    /// Суммирование функций
    /// Пример, где параллельные вычисления 
    /// имеют преимущество
    /// </summary>
    public class Sum_Function
    {
        double[] ar;    //массив
        int n; //Размер массива
        int p;  //Число групп при делении массива                
        double[] y;     // рабочий массив
        double S = 0;

         /// <summary>
        /// конструктор
        /// </summary>
        /// <param name="size"></param>
        /// <param name="p"></param>
        public Sum_Function(int size, int p)
        {
            n = size;
            this.p = p;
            ar = new double[n];
            y = new double[n];
        }
        public void Results(out double s, out double f)
        {
            s = S;
            f = ar[0] + ar[n - 1];
        }
        int Fx(int i)
        {
            //имитация длинных вычислений
            const int m = 100000;
            int add = 1, t = 0;
            for (int k = 0; k < m; k++)
            {
                t += add;
                add *= -1;
            }
            return i * 10;
        }
        int Fs(int i)
        {
            //имитация длинных вычислений
            const int m = 100000;
            int add = 1, t = 0;
            for (int k = 0; k < m; k++)
            {
                t += add;
                add *= -1;
            }
            return i;
        }

        /// <summary>
        /// Последовательный алгоритм суммирования функций
        /// В поле S сумма функций Fs(i)
        /// В массиве ar[i] - Fx(i)
        /// </summary>
        public void Seq_SumF()
        {
            S = 0;
            for (int i = 0; i < n; i++)
            {
                ar[i] = Fx(i);
                S = S + Fs(i);
            }
        }
        /// <summary>
        /// Некорректное применение метода Parallel.For
        /// </summary>
        public void UncorrectParallel()
        {
            S = 0;
            Parallel.For(0, n, (i) =>
            {
                ar[i] = Fx(i);
                S = S + Fs(i);
            });
        }

        /// <summary>
        /// Частичное распараллеливание
        /// Создаются два цикла - один распараллеливается
        /// </summary>
        public void Partition_Parallel()
        {
            S = 0;
            for (int i = 0; i < n; i++)
                S = S + Fs(i);
            Parallel.For(0, n, (i) =>
            {
                ar[i] = Fx(i);
            });
        }

        /// <summary>
        /// Полное распараллеливание
        /// вычисления функций
        /// Используется вспомогательный массив y
        /// </summary>
        public void Full_Parallel()
        {
            Parallel.For(0, n, (i) =>
            {
                ar[i] = Fx(i);
                y[i] = Fs(i);
            });
            S = 0;
            for (int i = 0; i < n; i++)
                S = S + y[i];
        }

        /// <summary>
        /// Шаговый параллельный алгоритм
        /// </summary>
        public void Step_SumF()
        {
            y = new double[p];
            Parallel.For(0, p, (i) =>
            {
                for (int j = i; j < n; j += p)
                {
                    ar[j] = Fx(j);
                    y[i] += Fs(j);
                }
            });
            S = y[0];
            for (int i = 1; i < p; i++)
                S += y[i];

        }

        /// <summary>
        /// параллельный алгоритм, 
        /// работающий непосредственно с потоками
        /// </summary>
        public void Parallel_with_Threads()
        {
            Thread[] threads = new Thread[n];
            y = new double[n];
            //создаем потоки
            for (int i = 0; i < n; i++)
            {
                threads[i] = new Thread(SumT);
            }
            //запускаем потоки
            for (int i = 0; i < n; i++)
                threads[i].Start(i);
            //Ждем завершения
            for (int i = 0; i < n; i++)
                threads[i].Join();
            //Продолжаем работу
            S = 0;
            for (int i = 0; i < n; i++)
                S = S + y[i];
        }
        /// <summary>
        /// Метод, вызываемый в потоке
        /// </summary>
        /// <param name="ind">номер потока</param>
        void SumT(object ind)
        {
            int k = (int)ind;
            ar[k] = Fx(k);
            y[k] = Fs(k);
        }
        /// <summary>
        /// Применяем задачи
        /// </summary>
        public void Parallel_with_Tasks()
        {
            Task[] tasks = new Task[n];
            y = new double[n];
            //создаем и запускаем задачи
            for (int i = 0; i < n; i++)
            {
                tasks[i] = new Task(SumT, i);
            }
            for (int i = 0; i < n; i++)
                tasks[i].Start();
            //Ждем завершения
            Task.WaitAll(tasks);
            //Продолжаем работу
            S = 0;
            for (int i = 0; i < n; i++)
                S = S + y[i];
        }

    }
}
